"""This module contains HeAT's version information."""

major: int = 1
"""Indicates HeAT's main version."""
minor: int = 1
"""Indicates feature extension."""
micro: int = 1
"""Indicates revisions for bugfixes."""
extension: str = None
"""Indicates special builds, e.g. for specific hardware."""

if not extension:
    __version__: str = "{}.{}.{}".format(major, minor, micro)
    """The combined version string, consisting out of major, minor, micro and possibly extension."""
else:
    __version__: str = "{}.{}.{}-{}".format(major, minor, micro, extension)



"""
Functions for relational oprations, i.e. equal/no equal...
"""
from __future__ import annotations

import torch

from typing import Union

from .communication import MPI
from .dndarray import DNDarray
from . import _operations
from . import dndarray
from . import types

__all__ = [
    "eq",
    "equal",
    "ge",
    "greater",
    "greater_equal",
    "gt",
    "le",
    "less",
    "less_equal",
    "lt",
    "ne",
    "not_equal",
]


def eq(x: Union[DNDarray, float, int], y: Union[DNDarray, float, int]) -> DNDarray:
    """
    Returns a :class:`~heat.core.dndarray.DNDarray` containing the results of element-wise comparision.
    Takes the first and second operand (scalar or :class:`~heat.core.dndarray.DNDarray`) whose elements are to be
    compared as argument.

    Parameters
    ----------
    x: DNDarray or scalar
        The first operand involved in the comparison
    y: DNDarray or scalar
        The second operand involved in the comparison

    Examples
    ---------
    >>> import heat as ht
    >>> x = ht.float32([[1, 2],[3, 4]])
    >>> ht.eq(x, 3.0)
    DNDarray([[False, False],
              [ True, False]], dtype=ht.bool, device=cpu:0, split=None)
    >>> y = ht.float32([[2, 2], [2, 2]])
    >>> ht.eq(x, y)
    DNDarray([[False,  True],
              [False, False]], dtype=ht.bool, device=cpu:0, split=None)
    """
    res = _operations.__binary_op(torch.eq, x, y)

    if res.dtype != types.bool:
        res = dndarray.DNDarray(
            res.larray.type(torch.bool),
            res.gshape,
            types.bool,
            res.split,
            res.device,
            res.comm,
            res.balanced,
        )

    return res


DNDarray.__eq__ = lambda self, other: eq(self, other)
DNDarray.__eq__.__doc__ = eq.__doc__


def equal(x: Union[DNDarray, float, int], y: Union[DNDarray, float, int]) -> bool:
    """
    Overall comparison of equality between two :class:`~heat.core.dndarray.DNDarray`. Returns ``True`` if two arrays
    have the same size and elements, and ``False`` otherwise.

    Parameters
    ----------
    x: DNDarray or scalar
        The first operand involved in the comparison
    y: DNDarray or scalar
        The second operand involved in the comparison

    Examples
    ---------
    >>> import heat as ht
    >>> x = ht.float32([[1, 2],[3, 4]])
    >>> ht.equal(x, ht.float32([[1, 2],[3, 4]]))
    True
    >>> y = ht.float32([[2, 2], [2, 2]])
    >>> ht.equal(x, y)
    False
    >>> ht.equal(x, 3.0)
    False
    """
    result_tensor = _operations.__binary_op(torch.equal, x, y)

    if result_tensor.larray.numel() == 1:
        result_value = result_tensor.larray.item()
    else:
        result_value = True

    return result_tensor.comm.allreduce(result_value, MPI.LAND)


def ge(x: Union[DNDarray, float, int], y: Union[DNDarray, float, int]) -> DNDarray:
    """
    Returns a D:class:`~heat.core.dndarray.DNDarray` containing the results of element-wise rich greater than or equal comparison between values from operand ``x`` with respect to values of
    operand ``y`` (i.e. ``x>=y``), not commutative. Takes the first and second operand (scalar or
    :class:`~heat.core.dndarray.DNDarray`) whose elements are to be compared as argument.

    Parameters
    ----------
    x: DNDarray or scalar
        The first operand to be compared greater than or equal to second operand
    y: DNDarray or scalar
       The second operand to be compared less than or equal to first operand

    Examples
    -------
    >>> import heat as ht
    >>> x = ht.float32([[1, 2],[3, 4]])
    >>> ht.ge(x, 3.0)
    DNDarray([[False, False],
              [ True,  True]], dtype=ht.bool, device=cpu:0, split=None)
    >>> y = ht.float32([[2, 2], [2, 2]])
    >>> ht.ge(x, y)
    DNDarray([[False,  True],
              [ True,  True]], dtype=ht.bool, device=cpu:0, split=None)
    """
    res = _operations.__binary_op(torch.ge, x, y)

    if res.dtype != types.bool:
        res = dndarray.DNDarray(
            res.larray.type(torch.bool),
            res.gshape,
            types.bool,
            res.split,
            res.device,
            res.comm,
            res.balanced,
        )

    return res


DNDarray.__ge__ = lambda self, other: ge(self, other)
DNDarray.__ge__.__doc__ = ge.__doc__

# alias
greater_equal = ge
greater_equal.__doc__ = ge.__doc__


def gt(x: Union[DNDarray, float, int], y: Union[DNDarray, float, int]) -> DNDarray:
    """
    Returns a :class:`~heat.core.dndarray.DNDarray` containing the results of element-wise rich greater than comparison between values from operand ``x`` with respect to values of
    operand ``y`` (i.e. ``x>y``), not commutative. Takes the first and second operand (scalar or
    :class:`~heat.core.dndarray.DNDarray`) whose elements are to be compared as argument.

    Parameters
    ----------
    x: DNDarray or scalar
       The first operand to be compared greater than second operand
    y: DNDarray or scalar
       The second operand to be compared less than first operand

    Examples
    -------
    >>> import heat as ht
    >>> x = ht.float32([[1, 2],[3, 4]])
    >>> ht.gt(x, 3.0)
    DNDarray([[False, False],
              [False,  True]], dtype=ht.bool, device=cpu:0, split=None)
    >>> y = ht.float32([[2, 2], [2, 2]])
    >>> ht.gt(x, y)
    DNDarray([[False, False],
              [ True,  True]], dtype=ht.bool, device=cpu:0, split=None)
    """
    res = _operations.__binary_op(torch.gt, x, y)

    if res.dtype != types.bool:
        res = dndarray.DNDarray(
            res.larray.type(torch.bool),
            res.gshape,
            types.bool,
            res.split,
            res.device,
            res.comm,
            res.balanced,
        )

    return res


DNDarray.__gt__ = lambda self, other: gt(self, other)
DNDarray.__gt__.__doc__ = gt.__doc__

# alias
greater = gt
greater.__doc__ = gt.__doc__


def le(x: Union[DNDarray, float, int], y: Union[DNDarray, float, int]) -> DNDarray:
    """
    Return a :class:`~heat.core.dndarray.DNDarray` containing the results of element-wise rich less than or equal comparison between values from operand ``x`` with respect to values of
    operand ``y`` (i.e. ``x<=y``), not commutative. Takes the first and second operand (scalar or
    :class:`~heat.core.dndarray.DNDarray`) whose elements are to be compared as argument.

    Parameters
    ----------
    x: DNDarray or scalar
       The first operand to be compared less than or equal to second operand
    y: DNDarray or scalar
       The second operand to be compared greater than or equal to first operand

    Examples
    -------
    >>> import heat as ht
    >>> x = ht.float32([[1, 2],[3, 4]])
    >>> ht.le(x, 3.0)
    DNDarray([[ True,  True],
              [ True, False]], dtype=ht.bool, device=cpu:0, split=None)
    >>> y = ht.float32([[2, 2], [2, 2]])
    >>> ht.le(x, y)
    DNDarray([[ True,  True],
              [False, False]], dtype=ht.bool, device=cpu:0, split=None)
    """
    res = _operations.__binary_op(torch.le, x, y)

    if res.dtype != types.bool:
        res = dndarray.DNDarray(
            res.larray.type(torch.bool),
            res.gshape,
            types.bool,
            res.split,
            res.device,
            res.comm,
            res.balanced,
        )

    return res


DNDarray.__le__ = lambda self, other: le(self, other)
DNDarray.__le__.__doc__ = le.__doc__

# alias
less_equal = le
less_equal.__doc__ = le.__doc__


def lt(x: Union[DNDarray, float, int], y: Union[DNDarray, float, int]) -> DNDarray:
    """
    Returns a :class:`~heat.core.dndarray.DNDarray` containing the results of element-wise rich less than comparison between values from operand ``x`` with respect to values of
    operand ``y`` (i.e. ``x<y``), not commutative. Takes the first and second operand (scalar or
    :class:`~heat.core.dndarray.DNDarray`) whose elements are to be compared as argument.

    Parameters
    ----------
    x: DNDarray or scalar
        The first operand to be compared less than second operand
    y: DNDarray or scalar
        The second operand to be compared greater than first operand

    Examples
    -------
    >>> import heat as ht
    >>> x = ht.float32([[1, 2],[3, 4]])
    >>> ht.lt(x, 3.0)
    DNDarray([[ True,  True],
              [False, False]], dtype=ht.bool, device=cpu:0, split=None)
    >>> y = ht.float32([[2, 2], [2, 2]])
    >>> ht.lt(x, y)
    DNDarray([[ True, False],
              [False, False]], dtype=ht.bool, device=cpu:0, split=None)
    """
    res = _operations.__binary_op(torch.lt, x, y)

    if res.dtype != types.bool:
        res = dndarray.DNDarray(
            res.larray.type(torch.bool),
            res.gshape,
            types.bool,
            res.split,
            res.device,
            res.comm,
            res.balanced,
        )

    return res


DNDarray.__lt__ = lambda self, other: lt(self, other)
DNDarray.__lt__.__doc__ = lt.__doc__

# alias
less = lt
less.__doc__ = lt.__doc__


def ne(x: Union[DNDarray, float, int], y: Union[DNDarray, float, int]) -> DNDarray:
    """
    Returns a :class:`~heat.core.dndarray.DNDarray` containing the results of element-wise rich comparison of non-equality between values from two operands, commutative.
    Takes the first and second operand (scalar or :class:`~heat.core.dndarray.DNDarray`) whose elements are to be
    compared as argument.

    Parameters
    ----------
    x: DNDarray or scalar
        The first operand involved in the comparison
    y: DNDarray or scalar
        The second operand involved in the comparison

    Examples
    ---------
    >>> import heat as ht
    >>> x = ht.float32([[1, 2],[3, 4]])
    >>> ht.ne(x, 3.0)
    DNDarray([[ True,  True],
              [False,  True]], dtype=ht.bool, device=cpu:0, split=None)
    >>> y = ht.float32([[2, 2], [2, 2]])
    >>> ht.ne(x, y)
    DNDarray([[ True, False],
              [ True,  True]], dtype=ht.bool, device=cpu:0, split=None)
    """
    res = _operations.__binary_op(torch.ne, x, y)

    if res.dtype != types.bool:
        res = dndarray.DNDarray(
            res.larray.type(torch.bool),
            res.gshape,
            types.bool,
            res.split,
            res.device,
            res.comm,
            res.balanced,
        )

    return res


DNDarray.__ne__ = lambda self, other: ne(self, other)
DNDarray.__ne__.__doc__ = ne.__doc__

# alias
not_equal = ne
not_equal.__doc__ = ne.__doc__



"""
Learning rate schedulers in the heat namespace
"""

import sys
import torch.optim.lr_scheduler as lrs

if sys.version_info.minor >= 7:

    def __getattr__(name):
        """
        Call the torch learning rate scheduler of a specified name
        """
        try:
            return lrs.__getattribute__(name)
        except AttributeError:
            raise AttributeError(f"name {name} is not implemented in torch.optim.lr_scheduler")


else:

    class _Wrapper36(object):
        """
        Class to add the torch learning rate schedulers to the heat.optim namespace
        """

        def __init__(self, wrapped):  # noqa: D107
            self.wrapped = wrapped

        def __getattr__(self, name):
            """
            Call the torch learning rate scheduler of a specified name
            """
            try:
                return lrs.__getattribute__(name)
            except AttributeError:
                raise AttributeError(f"name {name} is not implemented in torch.optim.lr_scheduler")

    sys.modules[__name__] = _Wrapper36(sys.modules[__name__])



"""
File with the available transforms for images
"""

import torchvision.transforms
import sys
import unittest


if sys.version_info.minor >= 7:

    def __getattr__(name):
        """
        When a function is called for the heat.vision_transforms module it will attempt to run the
        heat function/class with that name, then, if there is no such object, it will attempt to
        get the torchvision.transforms object of that name.
        """
        if name in torchvision.transforms.transforms.__all__:
            return torchvision.transforms.__getattribute__(name)
        else:
            raise AttributeError(f"module {name} not implemented in Torch or Heat")


else:

    class Wrapper(object):
        """
        Call through to torchvision.transforms
        """

        def __init__(self, wrapped):  # noqa: D107
            self.wrapped = wrapped

        def __getattr__(self, name):
            """
            When a function is called for the heat.vision_transforms module it will attempt to run the
            heat function/class with that name, then, if there is no such object, it will attempt to
            get the torchvision.transforms object of that name.
            """
            torch_all = torchvision.transforms.transforms.__all__
            if name in torch_all:
                return torchvision.transforms.__getattribute__(name)
            else:
                try:
                    unittest.__getattribute__(name)
                except AttributeError:
                    raise AttributeError(f"module {name} not implemented in Torch or Heat")

    sys.modules[__name__] = Wrapper(sys.modules[__name__])



"""
Collection of validation/sanitation routines.
"""
from __future__ import annotations

import numpy as np
import torch
import warnings
from typing import Any, Union, Sequence, List, Tuple

from .communication import MPI
from .dndarray import DNDarray

from . import factories
from . import stride_tricks
from . import types


__all__ = [
    "sanitize_in",
    "sanitize_infinity",
    "sanitize_in_tensor",
    "sanitize_lshape",
    "sanitize_out",
    "sanitize_sequence",
    "scalar_to_1d",
]


def sanitize_in(x: Any):
    """
    Verify that input object is ``DNDarray``.

    Parameters
    ----------
    x : Any
        Input object

    Raises
    ------
    TypeError
        When ``x`` is not a ``DNDarray``.
    """
    if not isinstance(x, DNDarray):
        raise TypeError("Input must be a DNDarray, is {}".format(type(x)))


def sanitize_infinity(x: Union[DNDarray, torch.Tensor]) -> Union[int, float]:
    """
    Returns largest possible value for the ``dtype`` of the input array.

    Parameters
    -----------
    x: Union[DNDarray, torch.Tensor]
        Input object.
    """
    dtype = x.dtype if isinstance(x, torch.Tensor) else x.larray.dtype
    try:
        largest = torch.finfo(dtype).max
    except TypeError:
        largest = torch.iinfo(dtype).max

    return largest


def sanitize_in_tensor(x: Any):
    """
    Verify that input object is ``torch.Tensor``.

    Parameters
    ----------
    x : Any
        Input object.

    Raises
    ------
    TypeError
        When ``x`` is not a ``torch.Tensor``.
    """
    if not isinstance(x, torch.Tensor):
        raise TypeError("Input must be a torch.Tensor, is {}".format(type(x)))


def sanitize_lshape(array: DNDarray, tensor: torch.Tensor):
    """
    Verify shape consistency when manipulating process-local arrays.

    Parameters
    ----------
    array : DNDarray
        the original, potentially distributed ``DNDarray``
    tensor : torch.Tensor
        process-local data meant to replace ``array.larray``

    Raises
    ------
    ValueError
        if shape of local ``torch.Tensor`` is inconsistent with global ``DNDarray``.
    """
    # input sanitation is done in parent function
    tshape = tuple(tensor.shape)
    if tshape == array.lshape:
        return
    gshape = array.gshape
    split = array.split
    if split is None:
        # allow for axes with size 0, other axes must match
        non_zero = list(i for i in range(len(tshape)) if tshape[i] != 0)
        cond = list(tshape[i] == gshape[i] for i in non_zero).count(True) == len(non_zero)
        if cond:
            return
        else:
            raise ValueError(
                "Shape of local tensor is inconsistent with global DNDarray: tensor.shape is {}, should be {}".format(
                    tshape, gshape
                )
            )
    # size of non-split dimensions must match global shape
    reduced_gshape = gshape[:split] + gshape[split + 1 :]
    reduced_tshape = tshape[:split] + tshape[split + 1 :]
    if reduced_tshape == reduced_gshape:
        return
    raise ValueError(
        "Shape of local tensor along non-split axes is inconsistent with global DNDarray: tensor.shape is {}, DNDarray is {}".format(
            tshape, gshape
        )
    )


def sanitize_out(out: Any, output_shape: Tuple, output_split: int, output_device: str):
    """
    Validate output buffer ``out``.

    Parameters
    ----------
    out : Any
          the `out` buffer where the result of some operation will be stored

    output_shape : Tuple
                   the calculated shape returned by the operation

    output_split : Int
                   the calculated split axis returned by the operation

    output_device : Str
                    "cpu" or "gpu" as per location of data

    Raises
    ------
    TypeError
        if ``out`` is not a ``DNDarray``.
    ValueError
        if shape, split direction, or device of the output buffer ``out`` do not match the operation result.
    """
    if not isinstance(out, DNDarray):
        raise TypeError("expected `out` to be None or a DNDarray, but was {}".format(type(out)))

    if out.gshape != output_shape:
        raise ValueError(
            "Expecting output buffer of shape {}, got {}".format(output_shape, out.shape)
        )
    if out.split is not output_split:
        raise ValueError(
            "Split axis of output buffer is inconsistent with split semantics (see documentation)."
        )
    if out.device is not output_device:
        raise ValueError(
            "Device mismatch: out is on {}, should be on {}".format(out.device, output_device)
        )


def sanitize_sequence(
    seq: Union[Sequence[int, ...], Sequence[float, ...], DNDarray, torch.Tensor]
) -> List:
    """
    Check if sequence is valid, return list.

    Parameters
    ----------
    seq : Union[Sequence[int, ...], Sequence[float, ...], DNDarray, torch.Tensor]
        Input sequence.

    Raises
    ------
    TypeError
        if ``seq`` is neither a list nor a tuple
    """
    if isinstance(seq, list):
        return seq
    elif isinstance(seq, tuple):
        return list(seq)
    else:
        raise TypeError("seq must be a list or a tuple, got {}".format(type(seq)))


def scalar_to_1d(x: DNDarray) -> DNDarray:
    """
    Turn a scalar ``DNDarray`` into a 1-D ``DNDarray`` with 1 element.

    Parameters
    ----------
    x : DNDarray
        with `x.ndim = 0`
    """
    return factories.array(
        x.larray.unsqueeze(0), dtype=x.dtype, split=x.split, comm=x.comm, device=x.device
    )
